"""Abstractions over dataclass fields."""
from __future__ import annotations

import dataclasses
from dataclasses import MISSING, fields
from typing import Any, Callable

from .types import ConfigField, Dataclass, UniversalKey

_METADATA_KEY = "pytoolconfig"


def field(
    default: Any = MISSING,
    description: str | None = None,
    command_line: tuple[str] | None = None,
    universal_config: UniversalKey | None = None,
    default_factory: Callable[[], Any] | None = MISSING,
    init: bool = True,
) -> dataclasses.Field:
    """Create a dataclass field with metadata."""
    metadata = {
        _METADATA_KEY: ConfigField(
            description=description,
            universal_config=universal_config,
            command_line=command_line,
            _default=default,
        )
    }

    if default_factory is not MISSING and default_factory is not None:
        metadata[_METADATA_KEY]._default = default_factory()
        return dataclasses.field(
            default_factory=default_factory, metadata=metadata, init=init
        )
    assert default is not MISSING
    return dataclasses.field(default=default, metadata=metadata, init=init)


def _gather_config_fields(
    model: type[Dataclass] | Dataclass,
) -> dict[str, ConfigField]:
    # First try PyToolConfig Annotated Fields
    result = {}
    for dataclass_field in fields(model):
        if dataclass_field.init:
            if _METADATA_KEY in dataclass_field.metadata:
                result[dataclass_field.name] = dataclass_field.metadata[_METADATA_KEY]
            else:
                result[dataclass_field.name] = ConfigField(
                    _default=dataclass_field.default
                )
            result[dataclass_field.name]._type = dataclass_field.type
    # Then use pydantic annotated fields
    if hasattr(model, "__pydantic_model__"):
        for pydantic_field in model.__pydantic_model__.__fields__.values():
            if pydantic_field.init:
                result[pydantic_field.name] = ConfigField(
                    description=pydantic_field.field_info.description,
                    _type=pydantic_field.type_,
                    _default=pydantic_field.default,
                )
                if "universal_config" in pydantic_field.field_info.extra:
                    result[
                        pydantic_field.name
                    ].universal_config = pydantic_field.field_info.extra[
                        "universal_config"
                    ]
                if "command_line" in pydantic_field.field_info.extra:
                    result[
                        pydantic_field.name
                    ].command_line = pydantic_field.field_info.extra["command_line"]
    return result



"""Tool to configure Python tools."""
from __future__ import annotations

from argparse import SUPPRESS, ArgumentParser
from dataclasses import is_dataclass
from pathlib import Path
from typing import Generic, Sequence, TypeVar

from pytoolconfig.fields import _gather_config_fields
from pytoolconfig.sources import PyProject, PyTool, Source
from pytoolconfig.types import ConfigField, Dataclass
from pytoolconfig.universal_config import UniversalConfig
from pytoolconfig.utils import _dict_to_dataclass, _recursive_merge

DataclassT = TypeVar("DataclassT", bound=Dataclass)


class PyToolConfig(Generic[DataclassT]):
    """Python Tool Configuration Aggregator."""

    sources: list[Source] = []
    tool: str
    working_directory: Path
    model: type[DataclassT]
    fall_through: bool = False
    arg_parser: ArgumentParser | None = None
    _config_fields: dict[str, ConfigField]

    def __init__(
        self,
        tool: str,
        working_directory: Path,
        model: type[DataclassT],
        arg_parser: ArgumentParser | None = None,
        custom_sources: Sequence[Source] | None = None,
        global_config: bool = False,
        global_sources: Sequence[Source] | None = None,
        fall_through: bool = False,
        *args,
        **kwargs,
    ):
        """Initialize the configuration object.

        :param tool: name of the tool to use.
        :param working_directory: working directory in use.
        :param model: Model of configuration.
        :param arg_parser: Arugument Parser.
        :param custom_sources: Custom sources
        :param global_config: Enable global configuration
        :param global_sources: Custom global sources
        :param fall_through: Configuration options should fall through between sources.
        :param args: Passed to constructor for PyProject
        :param kwargs: Passed to constructor for PyProject
        """
        assert is_dataclass(model)
        self.model = model
        self._config_fields = _gather_config_fields(model)
        self.tool = tool
        self.sources = [PyProject(working_directory, tool, *args, **kwargs)]
        if custom_sources:
            self.sources.extend(custom_sources)
        if global_config:
            self.sources.append(PyTool(tool))
        if global_sources:
            self.sources.extend(global_sources)

        self.arg_parser = arg_parser
        self.fall_through = fall_through
        self._setup_arg_parser()

    def parse(self, args: list[str] | None = None) -> DataclassT:
        """Parse the configuration.

        :param args: any additional command line overwrites.
        """
        configuration = self._parse_sources()
        assert isinstance(self.sources[0], PyProject)
        universal: UniversalConfig = self.sources[0].universalconfig()
        if self.arg_parser:
            if args is None:
                args = []
            parsed = self.arg_parser.parse_args(args)
            for name, value in parsed._get_kwargs():
                setattr(configuration, name, value)
        for name, field in self._config_fields.items():
            if field.universal_config:
                universal_value = vars(universal)[field.universal_config.name]
                if universal_value is not None:
                    setattr(
                        configuration,
                        name,
                        universal_value,
                    )
        return configuration

    def _setup_arg_parser(self) -> None:
        if self.arg_parser:
            for name, field in self._config_fields.items():
                if field.command_line:
                    flags = field.command_line
                    self.arg_parser.add_argument(
                        *flags,
                        type=field._type,
                        help=field.description,
                        default=SUPPRESS,
                        metavar=name,
                        dest=name,
                    )

    def _parse_sources(self) -> DataclassT:
        configuration = self.model()
        if self.fall_through:
            for source in reversed(self.sources):
                parsed = source.parse()
                if parsed is not None:
                    configuration = _recursive_merge(configuration, parsed)

        else:
            for source in self.sources:
                parsed = source.parse()
                if parsed:
                    return _dict_to_dataclass(self.model, parsed)
        return configuration



"""Version of pytoolconfig."""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from pytoolconfig import PyToolConfig, UniversalKey, field


@dataclass
class Version:
    version: str = field("UNKNOWN", universal_config=UniversalKey.version)


config = PyToolConfig("pytoolconfig", Path(__file__).parent, model=Version).parse()
version = config.version
